
# ---------------------------------------------------------------------
#  $Id: tkquery.tcl,v 1.70 1997/01/14 21:41:26 adabas Exp $
# ---------------------------------------------------------------------

# set some global defines...
set ROW_NOT_FOUND 100
set MASSCNT       20

# show objects only of current user...
set ownerCond "owner=USER"

# set what the Execute button should first do
set execCmd doSql

# Set default values for the option switches.
set options(dom:fontFamily)   {helvetica times courier}
set options(def:fontFamily)   helvetica
set options(dom:fontSize)     {8 10 12 14 18 24}
set options(def:fontSize)     14
set options(dom:fontStyle)    {roman bold italic bolditalic}
set options(def:fontStyle)    bold
set options(dom:autoCommit)   boolean
set options(def:autoCommit)   True
set options(dom:dateFormat)   {ansi eur internal iso jis oracle usa}
set options(def:dateFormat)   internal
set options(dom:outputStyle)  {plain html latex}
set options(def:outputStyle)  plain
set options(dom:withProt)     boolean
set options(def:withProt)     False
set options(dom:history)      integer
set options(def:history)      10
set options(dom:nullValue)    string
set options(def:nullValue)    ""
set options(dom:specialNull)  string
set options(def:specialNull)  "***"
set options(dom:withToolBar)  boolean
set options(def:withToolBar)  True
set options(dom:netscapeHelp) boolean
set options(def:netscapeHelp) False
set options(dom:balloonHelp)  boolean
set options(def:balloonHelp)  True
set options(dom:errorPopup)   boolean
set options(def:errorPopup)   False
# this corresponds to the following excerpt from ~/.Xdefaults...
# *tkquery.fontFamily:	 helvetica
# *tkquery.fontSize:	 14
# *tkquery.fontStyle:	 bold
# *tkquery.autoCommit:	 True
# *tkquery.dateFormat:	 internal
# *tkquery.outputStyle:	 plain
# *tkquery.withProt:	 False
# *tkquery.history:	 10
# *tkquery.nullValue:	
# *tkquery.specialNull:	 ***
# *tkquery.withToolBar:  True
# *tkquery.netscapeHelp: False
# *tkquery.balloonHelp:  True
# *tkquery.errorPopup:   False

frame .tkquery ;# Just for the options...

foreach optInd [array names options def:*] {
    set    optName  [string range $optInd 4 end]
    set    optClass [string toupper [string index $optName 0]]
    append optClass [string range $optName 1 end]

    option add *tkquery.$optName $options(def:$optName) widgetDefault
    set options(val:$optName) [option get .tkquery $optName $optClass]
    switch $options(dom:$optName) {
	boolean {
	    set options(val:$optName) [string match {[Tt]*} $options(val:$optName)]
	}
	integer {
	    if [catch {int($options(val:$optName))}] {
		set options(val:$optName) $options(def:$optName)
	    }
	}
	string {
	}
	default {
	    if {[lsearch -exact $options(dom:$optName) $options(val:$optName)]<0} {
		set options(val:$optName) $options(def:$optName)
	    }
	}
    }
}

set cur {}
set lda {}

set current(file)  ""
set current(query) ""
trace variable current w setHeader

set cmdIdx  0
set cmdLast 0
for {set i 0} {$i < $options(val:history)} {incr i} {
  set cmdRing($i) ""
}

set format(filSta,plain) ""
set format(filSta,html)  "
    <HTML><HEAD><TITLE>TkQuery Protocol</TITLE></HEAD>
    <BODY><H1>TkQuery Protocol</H1>"
set format(filSta,latex) "
\\documentclass\[german,a4paper]{article}
\\usepackage{babel}
\\pagestyle{empty}
\\begin{document}
\\small
"
set format(filEnd,plain) ""
set format(filEnd,html)  "</BODY>"
set format(filEnd,latex) "\\end{document}"
set format(header,plain) "headerPlain"
set format(header,html)  "headerHtml"
set format(header,latex) "headerLatex"
set format(vrbSta,plain) ""
set format(vrbSta,html)  "<PRE>"
set format(vrbSta,latex) "\\begin{verbatim}"
set format(vrbEnd,plain) ""
set format(vrbEnd,html)  "</PRE>"
set format(vrbEnd,latex) "\\end{verbatim}"
set format(hdrSta,plain) ""
set format(hdrSta,html)  "<TR><TH>"
set format(hdrSta,latex) "\\bf "
set format(hdrSep,plain) " "
set format(hdrSep,html)  "</TH><TH>"
set format(hdrSep,latex) "&\\bf "
set format(hdrEnd,plain) ""
set format(hdrEnd,html)  "</TH></TR><TR><TD>"
set format(hdrEnd,latex) "\\\\"
set format(colSep,plain) " "
set format(colSep,html)  "</TD><TD>"
set format(colSep,latex) "&"
set format(rowSep,plain) ""
set format(rowSep,html)  "</TD></TR><TR><TD>"
set format(rowSep,latex) "\\\\"
set format(filter,plain) "format"
set format(filter,html)  "filterHtml"
set format(filter,latex) "filterLatex"
set format(tailer,plain) "\n"
set format(tailer,html)  "</TD></TR></TABLE>"
set format(tailer,latex) "\\end{tabular}"


proc headerPlain {fmt cnt} {
    return ""
}

proc headerHtml {format cnt} {
    upvar $format fmt

    set result "<TABLE BORDER COLSPEC=\""
    for {set i 0} {$i < $cnt} {incr i} {
	append result [expr {[string match {%[0-9]*} $fmt($i)] ? "r " : "L "}]
    }
    append result "\">"
    return $result
}

proc headerLatex {format cnt} {
    upvar $format fmt

    set result "\\begin\{tabular\}\{"
    for {set i 0} {$i < $cnt} {incr i} {
	append result [expr {[string match {%[0-9]*} $fmt($i)] ? "r" : "l"}]
    }
    append result "\}"
    return $result
}

proc filterHtml {fmt data} {
    regsub -all {&} $data {\&amp;}   data
    regsub -all {} $data {\&auml;}  data
    regsub -all {} $data {\&ouml;}  data
    regsub -all {} $data {\&uuml;}  data
    regsub -all {} $data {\&Auml;}  data
    regsub -all {} $data {\&Ouml;}  data
    regsub -all {} $data {\&Uuml;}  data
    regsub -all {} $data {\&szlig;} data
    regsub -all {<} $data {\&lt;}    data
    regsub -all {>} $data {\&gt;}    data
    return $data
}

proc filterLatex {fmt data} {
    regsub -all {} $data "\"a"  data
    regsub -all {} $data "\"o"  data
    regsub -all {} $data "\"u"  data
    regsub -all {} $data "\"A"  data
    regsub -all {} $data "\"O"  data
    regsub -all {} $data "\"U"  data
    regsub -all {} $data "\"s"  data
    regsub -all {&} $data {\\\&} data
    return $data
}

########################
#
# handleList
#
#  called during a request for primary selection for listboxes and text
#  that bind the "selection get" command for button-2
#

proc handleList {w filtproc offset max} {
  set doproc [string length [info commands $filtproc]]
  if $doproc {
    set s "[$filtproc $w]"
  } else {
    # set s [$w get [lindex [$w curselection] 0] ]
    set i [$w curselection]  
    set n ""
    set s ""
    foreach r $i {
      set l [$w get $r] 
      append s "${n}$l"
      set n "\n"
    }

  } 
  return [string range $s $offset $max]
}

proc pannedListboxes {w} {
    global oldCols tkCompat

    pack [frame $w] -expand 1 -fill both
    set c [canvas $w.c -relief sunken -xscrollcommand "$w.horz set"]
    $c create window 1 1 -window [frame $c.f] -anchor nw -tag outframe
    grid rowconfigure $c.f 1 -weight $tkCompat(gridWeightMax)
    frame $c.f.pad
    set oldCols($c) 0
    bind $c <Configure> "canvasConf $c"

    label $w.pad -text ""
    if $tkCompat(fontCmd) {
	$w.pad configure -font outFont
    }

    scrollbar $w.vert -relief sunken -command "bindOutYView $c" \
	    -orient vertical -highlightthickness 0
    scrollbar $w.horz -relief sunken -command "$c xview" \
	    -orient horizontal -highlightthickness 0
    $w.vert set 0 1

    grid $c      -row 0 -column 0 -sticky nsew -rowspan 2
    grid $w.pad  -row 0 -column 1
    grid $w.vert -row 1 -column 1 -sticky ns
    grid $w.horz -row 2 -column 0 -sticky ew
    grid columnconfigure $w 0 -weight $tkCompat(gridWeightMax)
    grid rowconfigure    $w 1 -weight $tkCompat(gridWeightMax)
}

proc confListbox {w colNo action {colName ""}} {
    global oldCols tkCompat boxConf

    set c $w.c
    switch -- $action {
	<start> {
	    $c.f configure -width 10000
	}
	<stop> {
	    # forget old lists
	    for {set col $colNo} {$col < $oldCols($c)} {incr col} {
		foreach s [grid slaves $c.f -column $col] {
		    grid forget $s; destroy $s
		}
		grid columnconfigure $c.f $col -weight $tkCompat(gridWeightMin)
	    }

	    grid $c.f.pad -row 1 -column $colNo
	    grid columnconfigure $c.f $colNo -weight $tkCompat(gridWeightMax)

	    set oldCols($c) $colNo
	}
	<normal> - <disabled> {
	    for {set ix 0} {$ix < $oldCols($c)} {incr ix} {
		$c.f.l$ix configure -state [string trim $action <>]
	    }
	}
	default {
	    if {$colNo < $oldCols($c)} {
		$c.f.h$colNo configure $action $colName
		$c.f.l$colNo delete 1.0 end
	    } else {
		catch {destroy $c.f.h$colNo $c.f.l$colNo}
		set label [label $c.f.h$colNo $action $colName]
		set list  [text  $c.f.l$colNo -width $boxConf(width,$w,$colNo) \
			-highlightthickness 0 -wrap none -state disabled]
		if !$colNo {
		    $list configure -yscrollcommand "$w.vert set"
		}
		if ![string compare $boxConf(type,$w,$colNo) numeric] {
		    $list tag configure just -justify right
		}
		if $tkCompat(fontCmd) {
		    $label configure -font outFont
		    $list  configure -font outFont
		}
		grid $label -row 0 -column $colNo -ipadx 2 -padx 1 -sticky w
		grid $list  -row 1 -column $colNo -ipadx 2 -padx 1 -sticky nsew
	    }
	    grid columnconfigure $c.f $colNo -weight $tkCompat(gridWeightMin)
	    bind $c.f.h$colNo <3> "switchPanel $w $colNo visible 0"
	}
    }
}

proc setBoxConf {w colNo col} {
    global adamsg boxConf fmt

    set c       $w.c
    set headLen [string length $col]

    set boxConf(visible,$w,$colNo) 1
    set boxConf(title,$w,$colNo)   $col
    set boxConf(width,$w,$colNo)   $headLen

    switch -glob [lindex $adamsg(coltypes) $colNo] {
	long_byte* {
	    set fmt($colNo) "LBYTE"
	    set boxConf(type,$w,$colNo) data
	}
	long_* {
	    set fmt($colNo) "L%-50.50s"
	    set boxConf(type,$w,$colNo) text
	}
	fixed* - float* {
	    set fmt($colNo) "%[max [lindex $adamsg(collengths) $colNo] $headLen]s"
	    set boxConf(type,$w,$colNo) numeric
	}
	default {
	    set fmt($colNo) "%-[max [lindex $adamsg(collengths) $colNo] $headLen]s"
	    set boxConf(type,$w,$colNo) text
	}
    }
}

proc scrollBox {w dir} {
    switch $dir {
	left  {$w.c xview        scroll -1 pages}
	right {$w.c xview        scroll  1 pages}
	up    {bindOutYView $w.c scroll -1 pages}
	down  {bindOutYView $w.c scroll  1 pages}
	first {bindOutYView .o.c moveto  0}
	last  {bindOutYView .o.c moveto  1}
    }
}

proc insertListbox {w colNo {colValue ""}} {
    global boxConf

    set c $w.c
    set colWidth [string length $colValue]
    if {$colWidth > $boxConf(width,$w,$colNo)} {
	set boxConf(width,$w,$colNo) $colWidth
	$c.f.l$colNo configure -width $colWidth
    }
    $c.f.l$colNo insert end "$colValue\n" just
}

proc mkPanel {panel w} {
    global oldCols boxConf tkCompat

    if [catch {toplevel $panel}] {
	eval destroy [grid slaves $panel]
    }
    wm title $panel "Column Configure"

    grid [label $panel.l -text "Visible columns"] -row 0 -sticky ew

    set c $w.c
    for {set colNo 0} {$colNo < $oldCols($c)} {incr colNo} {
	set control [checkbutton $panel.l$colNo   \
		-textvariable boxConf(title,$w,$colNo)  \
		-variable     boxConf(visible,$w,$colNo) \
		-command  "switchPanel $w $colNo visible"]
	grid $control -row [expr {$colNo+1}] -column 0 -sticky w
    }

    grid [button $panel.b -text "Dismiss" -command "destroy $panel"] \
	    -row [expr {$colNo+2}] -column 0 -sticky ew
    grid columnconfigure $panel 0 -weight $tkCompat(gridWeightMax)

    update idletask
}

proc switchPanel {w colNo kind {value ""}} {
    global boxConf oldCols

    set c $w.c
    if [string length $value] {
	set boxConf($kind,$w,$colNo) $value
    }

    switch $kind {
	visible {
	    if $boxConf($kind,$w,$colNo) {
		grid $c.f.h$colNo -row 0 -column $colNo -ipadx 2 -padx 1 -sticky w
		grid $c.f.l$colNo -row 1 -column $colNo -ipadx 2 -padx 1 -sticky nsew
	    } else {
		grid forget $c.f.h$colNo $c.f.l$colNo
	    }
	}
	default {
	    return
	}
    }
    updateListboxes $c $oldCols($c)
}

proc updateListboxes {c colNo} {
    set bbox0 [grid bbox $c.f 0      1]
    set bboxN [grid bbox $c.f $colNo 1]
    set boxWidth [expr {[lindex $bboxN 0] - [lindex $bbox0 0]}]
    set infinity [$c.f cget -width]

    $c.f configure -width $boxWidth
    $c   configure -scrollregion [$c bbox outframe]
    $c.f configure -width $infinity
}

proc listboxGetLine {w mx} {
    global oldCols boxConf
    
    set c $w.c
    set line [$c.f.l$mx index current]
    for {set ix 0} {$ix < $oldCols($c)} {incr ix} {
	lappend res [$c.f.l$ix get "$line linestart" "$line lineend"]
    }
    return  $res
}

########################
#
# filtCols
#
#  a special filtproc to return a column list
#

proc filtCols {w} {
  set i [$w curselection]  
  set n ""
  set s ""
  foreach r $i {
    set l [$w get $r] 
    set l [eval list $l]
    append s "${n}[lindex $l 0]"
    set n ", "
  }
  return "$s"
}

########################
#
# filtTab
#
#  a special filtproc to return word 1 "." word 2
#

proc filtTab {w} {
  set s [$w get [lindex [$w curselection] 0] ]
  regsub -all \\$ $s \\\\$ s
  set s [eval list $s]
  return [string trim [lindex $s 0].[lindex $s 1]]
}

#
# createMain
#
#   create the main window
#

proc createMain {} {
    global lda cur adamsg connect options tcl_platform menu options tkCompat

    wm geom     . 600x500
    wm minsize  . 400 370
    wm title    . "Adabas D: TkQuery ($connect(serverdb)/$connect(user))"
    wm iconname . "TkQuery"

    # create a menu bar with some menu buttons
    if $tkCompat(menubars) {
	. configure -menu [menu .mb]

	set menu(sql)  .mb.sql
	set menu(exec) .mb.exec
	set menu(out)  .mb.out
	set menu(op)   .mb.op
	set menu(ob)   .mb.ob
	set menu(help) .mb.help

	.mb add cascade -label "SQL"         -menu $menu(sql)  -und 0
	.mb add cascade -label "Execute"     -menu $menu(exec) -und 0
	.mb add cascade -label "Results"     -menu $menu(out)  -und 0
	.mb add cascade -label "Options"     -menu $menu(op)   -und 0
	.mb add cascade -label "DB Objects"  -menu $menu(ob)   -und 0
	.mb add cascade -label "Help"        -menu $menu(help) -und 0
    } else {
	frame .mb -relief raised -borderwidth 2
	pack  .mb  -side top -fill x

	set menu(sql)  .mb.sql.m
	set menu(exec) .mb.exec.m
	set menu(out)  .mb.out.m
	set menu(op)   .mb.op.m
	set menu(ob)   .mb.ob.m
	set menu(help) .mb.help.m

	menubutton .mb.sql  -text "SQL"         -menu $menu(sql)  -und 0
	menubutton .mb.exec -text "Execute"     -menu $menu(exec) -und 0
	menubutton .mb.out  -text "Results"     -menu $menu(out)  -und 0
	menubutton .mb.op   -text "Options"     -menu $menu(op)   -und 0
	menubutton .mb.ob   -text "DB Objects"  -menu $menu(ob)   -und 0
	menubutton .mb.help -text "Help"        -menu $menu(help) -und 0

	pack .mb.sql .mb.exec .mb.out .mb.op .mb.ob -side left
	if ![string compare $tcl_platform(platform) unix] {
	    pack .mb.help -side right
	} else {
	    pack .mb.help -side left
	}
    }

    menu $menu(sql)
    $menu(sql) add command -label "New"  -command doNew        -und 2 -acc F2
    $menu(sql) add command -label "Prev" -command "prevSql -1" -und 0 -acc F9
    $menu(sql) add command -label "Next" -command "prevSql  1" -und 0 -acc F6
    $menu(sql) add separator
    $menu(sql) add command -label "Stored commands" -state disabled
    $menu(sql) add command -label "Get..."     -und 0 -command getQuery
    $menu(sql) add command -label "Replace"    -und 0 -command putQuery
    $menu(sql) add command -label "Insert..."  -und 0 -command putQueryAs
    $menu(sql) add command -label "Delete..."  -und 0 -command deleteQuery
    $menu(sql) add separator
    $menu(sql) add command -label "Open..."    -und 0 -command openSql
    $menu(sql) add command -label "Save"       -und 0 -command saveSql 
    $menu(sql) add command -label "Save As..." -und 5 -command saveSqlAs
    $menu(sql) add separator
    $menu(sql) add command -label "Exit" -command confirmExit -und 0 -acc F3

    menu $menu(exec)
    $menu(exec) add command -label "Execute"  -command doSql     -und 0 -acc F5
    $menu(exec) add command -label "Continue" -command fetchRows -und 1 \
                                              -state disabled
    $menu(exec) add command -label "Cancel"   -command doCancel  -und 0 -acc F3 \
                                              -state disabled
    $menu(exec) add sep
    $menu(exec) add command -label "Commit now" -underline 7     \
	    -command {adacommit $lda; setMsg "Committed" }
    $menu(exec) add command -label "Rollback now"   -underline 0 \
	    -command {adarollback $lda; setMsg "Rolled back" }

    menu $menu(out)
    $menu(out) add command -label "Clear" -command "clearoutput .o" -und 0
    $menu(out) add command -label "Configure..." -state disabled \
	    -command "mkPanel .panel .o"

    $menu(out) add separator
    $menu(out) add checkbutton -label "Protocol" -und 0 \
	    -var options(val:withProt) -command setProt
    $menu(out) add command -label "Save As..." -und 0 -command saveOut
    $menu(out) add command -label "Preview"    -und 3 -command preview
    $menu(out) add separator
    $menu(out) add command -label "Scroll Left"  -und 7 -acc F10 \
	    -command "scrollBox .o left"
    $menu(out) add command -label "Scroll Right" -und 7 -acc F11 \
	    -command "scrollBox .o right"

    menu $menu(op)
    $menu(op) add checkbutton -label "Autocommit" -underline 0 \
	    -variable options(val:autoCommit) -command setAutoCommit

    $menu(op) add command -label "Null Value..."   -und 0 -command setNull
    $menu(op) add command -label "Special Null..." -und 0 -command setSpecialNull
    $menu(op) add cascade -label "Date Format"     -und 0 -menu $menu(op).df
    menu $menu(op).df
    foreach format {ansi eur internal iso jis oracle usa} {
	$menu(op).df add radiobutton -variable options(val:dateFormat) \
		-value $format -label $format -command "setFormat"
    }
    $menu(op) add sep
    $menu(op) add checkbutton -label "Tool bar" -variable options(val:withToolBar) \
	    -command toggleToolBar
    $menu(op) add command -label "Font..." -und 0 -command "chooseFont .o.c"
    $menu(op) add checkbutton -label "Popup on error" -underline 9 \
	    -variable options(val:errorPopup)
    $menu(op) add cascade -label "Output Style" -und 0 -menu $menu(op).os
    menu $menu(op).os
    foreach style {plain html latex} {
	$menu(op).os add radiobutton -variable options(val:outputStyle) \
		-value $style -label $style -command setStyle
    }

    menu $menu(ob)
    $menu(ob) add command -label "Tables"  -command showTables  -underline 0
    $menu(ob) add command -label "Views"   -command showViews   -underline 0
    $menu(ob) add command -label "Indexes" -command showIndexes -underline 0
    $menu(ob) add sep
    $menu(ob) add command -label "Objects of..." -state disabled
    $menu(ob) add radio   -label "All"  -var ownerCond -value "1=1"
    $menu(ob) add radio   -label "User" -var ownerCond -value "owner=USER"

    menu $menu(help)
    $menu(help) add command -label "About..."     -und 0 -command aboutHelp
    $menu(help) add command -label "TkQuery..."   -und 2 -acc F1\
	    -command "callHtmlHelp tkquery"
    $menu(help) add command -label "AdabasTcl..." -und 6 \
	    -command "callHtmlHelp adabastcl"
    $menu(help) add check   -label "With balloon help" -und 5 \
	    -variable options(val:balloonHelp)
    $menu(help) add check   -label "Use Netscape" -und 4 \
	    -variable options(val:netscapeHelp)

    # create a message at the bottom (its important to pack it first,
    # since then it remains visible, even if window got very small...
    frame .msgBar
    label .msg -text "" -relief sunken -bd 1 -anchor w
    label .msgBar.foo -width 8 -relief sunken -bd 1
    pack  .msg -in .msgBar -side left -padx 2 -expand yes -fill both
    pack  .msgBar.foo -side left -padx 2
    pack  .msgBar -side bottom -fill x -pady 2
    if ![string compare $tcl_platform(platform) unix] {
	if $tkCompat(fontCmd) {
	    .msg configure -font "Helvetica 12 normal"
	} else {
	    .msg configure -font -*-Helvetica-Medium-R-Normal--*-120-*-*-*-*-*-*
	}
    }

    # create a button bar, if wanted.
    if $options(val:withToolBar) {
	toolBox .b
    }

    # Lets create the fonts for the sql and output windows
    setFont

    # create a frame listing sql code
    frame .s -relief raised -borderwidth 2

    label .s.l
    setHeader current "" w
    scrollbar .s.vert -relief sunken -command ".s.sql yview" \
	    -orient vertical -highlightthickness 0
    text .s.sql -relief sunken -height 8 -width 80 -yscroll ".s.vert set" \
	    -wrap word -highlightthickness 0 -bg white -foreground black
    if $tkCompat(fontCmd) {
	.s.sql configure -font fixedFont
    } else {
	catch {.s.sql configure -font -*-courier-*-r-*-*-14-*-*-*-*-*-*-*}
    }

    # bind some shortcut keys to the text window
    bind .s.sql <F5>    {if {$execCmd == "doSql"} $execCmd; break}
    bind .s.sql <F3>    {doCancel;             break}
    bind .s.sql <F1>    {callHtmlHelp tkquery; break}
    bind .s.sql <F2>    {doNew;                break}
    bind .s.sql <F9>    {prevSql -1;           break}
    bind .s.sql <F6>    {prevSql  1;           break}
    bind .s.sql <F10>   {scrollBox .o left;    break}
    bind .s.sql <F11>   {scrollBox .o right;   break}
    bind .s.sql <Prior> {scrollBox .o up;      break}
    bind .s.sql <Next>  {scrollBox .o down;    break}
    bind .s.sql <Home>  {scrollBox .o first;   break}
    bind .s.sql <End>   {scrollBox .o last;    break}
    bind .s.sql <F4>    {doSwitch;             break}

    # bind . <Enter> {focus .s.sql} ;# Itr seems, that we don't need this

    pack .s.l    -side top   -fill x
    pack .s.vert -side right -fill y
    pack .s.sql  -side left  -fill both -expand 1

    pack .s -side top -fill both

    pannedListboxes .o

    focus .s.sql

    # Initialize the options...
    initOptions .o.c

    # Now find out, if we are connected to a database server prior to 6.2.
    adasql $cur "SHOW (version) VERSION"
    adasql $cur "SELECT kernel INTO :knlVersion FROM version"
    adarollback $lda
    regexp { ([0-9])\.([0-9])} $knlVersion match major minor
    set connect(knlVersion) $major$minor

    # Remember some information, that depends on the kernel version.
    if {$connect(knlVersion) >= 62} {
	set connect(querylib)     querylib
	set connect(querylibrary) querylibrary
	set connect(query_own)    query_own
	set connect(query_call)   query_call
	set connect(indcoluse)    indcoluse
	set connect(blank_key)    blk_key
    } else {
	set connect(querylib)     sysquerylib
	set connect(querylibrary) sysquerylibrary
	set connect(query_own)    sysquery_own
	set connect(query_call)   sysquery_call
	set connect(indcoluse)    ind_uses_col
	set connect(blank_key)    blank_key
    }

    wm deiconify .
    setMsg
}

proc toggleToolBar {} {
    global options
    if $options(val:withToolBar) {
	toolBox .b -before .s
    } else {
	destroy .b
    }
}

proc callHtmlHelp {name} {
    global adabastcl_library options

    htmlHelp $adabastcl_library/manuals/$name.html $options(val:netscapeHelp)
}

proc canvasConf {c} {
    global oldCols

    if [grid propagate $c.f] {grid propagate $c.f 0}
    $c.f configure -height [winfo height $c]
    if $oldCols($c) {updateListboxes $c $oldCols($c)}
}

proc bindOutYView {c args} {
    global oldCols

    if [catch "$c.f.l0 yview $args"] return
    set topLine [$c.f.l0 index @0,0]
    for {set ix 1} {$ix < $oldCols($c)} {incr ix} {
	$c.f.l$ix yview $topLine
    }
}

proc chooseFont {c} {
    global options

    toplevel .font
    wm title .font "Font for results"

    label .font.fontFamily -text "Family"
    grid  .font.fontFamily -row [set row 0] -column 0
    foreach name $options(dom:fontFamily) {
	radiobutton .font.$name -variable options(val:fontFamily) \
		-value $name -text $name -command "setFont $c"
	grid .font.$name -row [incr row] -column 0 -sticky w
    }

    label .font.fontSize -text "Size"
    grid  .font.fontSize -row [set row 0] -column 1
    foreach size $options(dom:fontSize) {
	radiobutton .font.s$size -variable options(val:fontSize) \
		-value $size -text $size -command "setFont $c"
	grid .font.s$size -row [incr row] -column 1 -sticky w
    }
    
    label .font.fontStyle -text "Style"
    grid  .font.fontStyle -row [set row 0] -column 2
    foreach style $options(dom:fontStyle) {
	radiobutton .font.$style -variable options(val:fontStyle) \
		-value $style -text $style -command "setFont $c"
	grid .font.$style -row [incr row] -column 2 -sticky w
    }

    button .font.dismiss -text Dismiss -command "destroy .font"
    grid   .font.dismiss - - -column 0 -sticky ew

    bind .font <Return> ".font.dismiss invoke"
}

if $tkCompat(fontCmd) {
    proc setFont {{c ""}} {
	global options

	switch $options(val:fontStyle) {
	    roman      {set weight normal; set slant roman}
	    bold       {set weight bold;   set slant roman}
	    italic     {set weight normal; set slant italic}
	    bolditalic {set weight bold;   set slant italic}
	}
	if [string length $c] {
	    set kind configure
	} else {
	    set kind create
	}
	font $kind outFont \
		-family $options(val:fontFamily) \
		-size   $options(val:fontSize)   \
		-weight $weight \
		-slant  $slant
	font $kind fixedFont \
		-family Courier \
		-size   $options(val:fontSize)   \
		-weight $weight \
		-slant  $slant

	if [string length $c] {
	    canvasConf $c
	}
	update idletask
    }
} else {
    proc setFont {{c ""}} {
	global options font

	switch $options(val:fontStyle) {
	    roman      {set weight medium; set style r}
	    bold       {set weight bold;   set style r}
	    italic     {set weight medium; set style o}
	    bolditalic {set weight bold;   set style o}
	}
	# For times it is called italic instead of oblique...
	if {![string compare $options(val:fontFamily) times] \
		&& ![string compare $style o]} {
	    set style i
	}

	set font -*-$options(val:fontFamily)-$weight-$style-*-*-$options(val:fontSize)-*-*-*-*-*-*-*
	if [string length $c] {
	    if [catch {.o.pad configure -font $font}] return
	    set n 0
	    while 1 {
		if [catch {$c.f.h$n configure -font $font} msg] break
		if [catch {$c.f.l$n configure -font $font} msg] break
		incr n
	    }
	    canvasConf $c
	}
	option add *o.c*font $font interactive
	update idletask
    }
}

proc setStyle {{occasion onRadio}} {
    global options format protFilename oldStyle

    if ![catch {open $protFilename a} f] {
	if ![info exists oldStyle] {
	    set oldStyle $options(val:outputStyle)
	}
	puts  $f $format(filEnd,$oldStyle)
	close $f
    }
    set oldStyle $options(val:outputStyle)

    if ![string compare $occasion onRadio] {
	setProt
    }
}

proc setProt {} {
    global options format protFilename errorCode

    if $options(val:withProt) {
	switch $options(val:outputStyle) {
	    html  {set protFilename "tkquery.html"}
	    latex {set protFilename "tkquery.tex"}
	    plain {set protFilename "tkquery.prot"}
	}
	if [catch {open $protFilename w} f] {
	    setMsg "Error while open '$protFilename':\
[lindex $errorCode 2], no protocol"
	    set options(val:withProt) 0
	} else {
	    puts  $f $format(filSta,$options(val:outputStyle))
	    close $f
	}
    }
}

proc doCancel {} {
    global execCmd

    if [string compare $execCmd doSql] {
	uplevel #0 $execCmd
    } else {
	confirmExit
    }
}


proc setAutoCommit {{noMsg 0}} {
    global lda options menu

    set onOff [expr {$options(val:autoCommit) ? "on" : "off"}]
    adaautocom $lda $onOff;
    if !$noMsg {
	setMsg "Autocommit $onOff"
    }

    set state [expr {$options(val:autoCommit) ? "disabled" : "normal"}]
    foreach entry {"Commit now" "Rollback now*"} {
	confMenueEntry $menu(exec)  $entry -state $state
    }
}

proc setFormat {{noMsg 0}} {
    global options cur adamsg connect

    if [string compare $connect(sqlmode) oracle] {
	if [catch {callSql $cur "SET FORMAT $options(val:dateFormat)" \
		-sqlmode adabas}] {
	    set msg "Error $adamsg(rc):$adamsg(errortxt) while setting format" 
	} else {
	    set msg "Set format $options(val:dateFormat)"
	}
	if !$noMsg {
	    setMsg $msg
	}
    }
}

#
# setMsg
#
#   set the text for the label at bottom of results window
#

proc setMsg {args}  {
    global options defaultMsg

    set popup    0

    switch [llength $args] {
	0 {set msg [set defaultMsg ""]}
	1 {
	    if ![string compare -reset [lindex $args 0]] {
		catch {set defaultMsg} msg
	    } else {
		set msg   [set defaultMsg [lindex $args 0]]
		set popup [expr {[regexp ^Error $msg] && $options(val:errorPopup)}]
	    }
	}
	2 {
	    if ![string compare -help [lindex $args 0]] {
		set msg [lindex $args 1]
	    }
	}
    }
    if ![info exists msg] {return -code error "usage: setMsg ?-help? msg"}

    if [regexp ^Error $msg] {
	.msg configure -text $msg -foreground red
    } else {
	set defaultColor [lindex [.msg configure -foreground] 3]
	.msg configure -text $msg -foreground $defaultColor
    }
    update idletasks

    if $popup {
	tk_messageBox -title "SQL Error" -icon error -message $msg
    }
}

proc setHeader {name1 name2 op} {
    upvar #0 $name1 current

    set textList ""
    if [string length $current(query)] {
	lappend textList "Query $current(query)"
    }
    if [string length $current(file)] {
	lappend textList "File $current(file)"
    }
    if [llength $textList] {
	set text [join $textList " / "]
    } else {
	set text noname
    }
    .s.l configure -text "SQL ($text)"
}

proc confirmExit {} {
  global lda cur

    switch [tk_messageBox -title "Exit?" -icon question \
	    -message "Really Exit?" -type yesno -default yes] {
	yes {
	    adaclose  $cur  ;# close the cursor,
	    adalogoff $lda  ;# disconnect from the database server,
	    setStyle onExit ;# finish the protocol,
	    exit            ;# and bye, bye...
	}
    }
}

#
# clearsql
#
#   clear the sql code window
#
proc clearsql {} {
  global current cmdIdx cmdLast

  set cmdIdx $cmdLast

  .s.sql delete 1.0 end
  set current(file)  ""
  set current(query) ""
  focus .s.sql
}

# clearoutput
#
#   clear the output canvas with all the listboxes
#
proc clearoutput {w {onlyReset 0}} {
    global menu

    if !$onlyReset {
	# Since there are no select columns, there is nothing to configure...
	confMenueEntry $menu(out)  "Configure..." -state disabled
	catch "destroy .panel"
    }

    confListbox $w 0 <stop>
    $w.c configure -scrollregion "0 0 0 0"
    $w.vert set 0 1

    focus .s.sql
}

proc topEntry {w title text varName {width 30}} {
    global _entryOk

    set _entryOk 0

    catch {destroy $w}
    toplevel $w -class Dialog
    makeTransient $w
    wm title $w $title

    frame $w.f1 -relief sunken -borderwidth 1

    label $w.f1.l -text $text
    entry $w.f1.e -width 18 -textvariable $varName -relief sunken \
	    -highlightthickness 0 -background white -foreground black

    pack $w.f1.l -side left
    pack $w.f1.e -side left -fill x -expand 1

    frame $w.f3 -relief sunken -borderwidth 1
    button $w.f3.can -text "Cancel" -command "destroy $w"
    button $w.f3.app -text "Set"    -command "set _entryOk 1; destroy $w"

    pack $w.f3.app $w.f3.can -side left -expand 1 -fill x
    pack $w.f1 $w.f3 -side top -padx 10 -pady 5 -fill both

    bind  $w.f1.e <F3>     "$w.f3.can invoke; break"
    bind  $w.f1.e <Return> "$w.f3.app invoke; break"

    grab  $w
    focus $w.f1.e
    tkwait window $w

    return $_entryOk
}

# ---------------------------------------------------------------------------
# selectQuery: pops up a toplevel with a listbox, to let the user choose
#              a stored command by selecting its name.
# ---------------------------------------------------------------------------
proc selectQuery {title allQueries} {
    global cur _queryName connect tkCompat

    catch {destroy .queries}
    toplevel .queries
    wm title .queries $title

    listbox   .queries.l           -yscrollcommand ".queries.s set"
    scrollbar .queries.s -orient vertical -command ".queries.l yview"
    frame     .queries.b
    button    .queries.b.dis -text Dismiss -command {destroy .queries}
    button    .queries.b.ok  -text Okay    -command {
	set _queryName [.queries.l get active]; destroy .queries
    }

    grid .queries.l -row 0 -column 0 -sticky nsew
    grid .queries.s -row 0 -column 1 -sticky ns
    grid .queries.b -row 1 -column 0 -sticky ew -columnspan 2

    grid    rowconfigure .queries 0 -weight $tkCompat(gridWeightMax)
    grid columnconfigure .queries 0 -weight $tkCompat(gridWeightMax)

    grid .queries.b.dis -row 0 -column 0 -sticky ew
    grid .queries.b.ok  -row 0 -column 1 -sticky ew
    grid columnconfigure .queries.b 0 -weight $tkCompat(gridWeightMax)
    grid columnconfigure .queries.b 1 -weight $tkCompat(gridWeightMax)

    bind .queries.l <F3>       ".queries.b.dis invoke; break"
    bind .queries.l <Return>   ".queries.b.ok  invoke; break"
    bind .queries.l <Double-1> ".queries.b.ok  invoke; break"

    set queryFound 0
    set select "SELECT command_name
                  FROM $connect(query_own)
                  ORDER BY command_name"
    if ![catch {callSql $cur $select -sqlmode adabas}] {
	set queryFound 1
	adafetch $cur -command ".queries.l insert end @1"
    }
    if $allQueries {
	set select "SELECT author || '.' || command_name
                      FROM $connect(query_call)
                      ORDER BY author, command_name"
	if ![catch {callSql $cur $select -sqlmode adabas}] {
	    set queryFound 1
	    adafetch $cur -command ".queries.l insert end @1"
	}
    }
    if !$queryFound {
	setMsg "No stored query found"
	return
    }

    focus .queries.l
    tkwait window .queries

    # If the user pressed Dismiss or destroyed the window, better jump off.
    if ![info exists _queryName] {return ""}

    # Make sure, the _queryName isn't defined after leaving the proc.
    set localName $_queryName
    unset _queryName
    return $localName
}

proc getQuery {} {
    global cur current connect

    setMsg
    set queryName [selectQuery "Get query" 1]
    if ![string length $queryName] return 

    if [regexp {^([^.]*)\.(.*)$} $queryName match author queryName] {
	set queryLib SYSDBA($author).$connect(querylibrary)
	set author   '$author'
    } else {
	set queryLib $connect(querylib)
	set author   USERGROUP
    }
    set select "SELECT command
                  FROM $queryLib
                  WHERE author       = $author
                    AND command_name = '$queryName'"
    callSql $cur $select -sqlmode adabas
    set cmd [adareadlong $cur -descriptor [lindex [adafetch $cur] 0]]
    set cmdLines [join [adabas get $cmd editform] \n]

    clearsql
    clearoutput .o
    .s.sql insert 1.0 $cmdLines

    setMsg "Query $queryName loaded"
    set current(query) $queryName
}

proc putQuery {{kind replaced}} {
    global current cur lda connect

    setMsg
    if ![string length $current(query)] {
	return [putQueryAs]
    }

    set  cmdString [.s.sql get 1.0 end]
    set  strLength [string length $cmdString]
    set  cmdList   [split $cmdString "\n"]
    incr strLength [llength $cmdList]

    set whereCond "author = USER AND objtype = 'QUERYCOMMAND' \
                   AND command_name = '$current(query)'"
    set select "UPDATE $connect(querylib)
                  SET updatedate = DATE,
                      updatetime = TIME,
                      strlength  = $strLength
                  WHERE $whereCond"
    if [catch {adasql $cur $select} msg] {
	if ![string compare $msg 100] {
	    return [putQueryAs]
	}
	setMsg "Error: $msg"
	return
    }

    adawritelong $cur -table $connect(querylib) -column command \
                      -where $whereCond -value [adabas put $cmdList editform]
    adacommit $lda

    setMsg "Command $current(query) $kind."
}

proc putQueryAs {} {
    global current cur _queryName connect

    setMsg
    set _queryName $current(query)

    if {![topEntry .storeAs "Store Query As" "Query name" _queryName 18] \
	    || ![string length $_queryName]} {
	unset _queryName
	return
    }

    # The insert doesn't like it at all, if command_name is too long,
    # and xquery doesn't allow a dot...
    set _queryName [string range $_queryName 0 17]
    regsub -all "\\." $_queryName ":" _queryName

    set insert "INSERT $connect(querylib) (command_name, codetype,
                                 $connect(blank_key)4, $connect(blank_key)5, \
                                 $connect(blank_key)6, $connect(blank_key)7) \
                         VALUES ('$_queryName', 'ASCII', '', '', '', '')"

    if [catch {callSql $cur $insert -sqlmode adabas} ret] {

	if {$ret != 200} {return -code error $ret}

	switch [tk_messageBox -title "Replace query?" \
		-message "Query $_queryName already exists! Replace it?" \
		-icon question -default cancel -type okcancel] {
	    cancel {return}
	}

	set current(query) $_queryName
	return [putQuery]
    }

    set insert "INSERT $connect(query_own) (command_name, $connect(blank_key)2, \
                                  $connect(blank_key)3,  $connect(blank_key)5,  \
                                  $connect(blank_key)6,  $connect(blank_key)7,  \
                                  $connect(blank_key)11, $connect(blank_key)12, \
                                  $connect(blank_key)13, $connect(blank_key)14) \
                          VALUES ('$_queryName', '', '', '',                    \
                                  '', '', '', '', '', '')"
    callSql $cur $insert -sqlmode adabas

    set current(query) $_queryName
    unset _queryName
    putQuery inserted    
}


proc deleteQuery {} {
    global cur lda current connect

    setMsg
    set queryName [selectQuery "Delete query" 0]
    if ![string length $queryName] return 

    callSql $cur "DELETE FROM $connect(querylib)        \
                    WHERE author       = USER           \
                      AND objtype      = 'QUERYCOMMAND' \
                      AND command_name = '$queryName'" -sqlmode adabas
    callSql $cur "DELETE FROM $connect(query_own)       \
                    WHERE author       = USER           \
                      AND refobjtype   = 'QUERYCOMMAND' \
                      AND command_name = '$queryName'" -sqlmode adabas
    adacommit $lda

    setMsg "Query $queryName deleted"
    if ![string compare $current(query) $queryName] {
	set current(query) ""
    }
}

#
# openSql
#
#    try to open the file selected by the user in a file selection box

proc openSql {} {
    global current cmdIdx cmdLast

    set fileName [tk_getOpenFile -title "TkQuery: Open SQL File" \
	    -filetypes {{"SQL command file" *.sql} {"All Files" *}}]
    if ![string length $fileName] return

    set cmdIdx $cmdLast

    if [file isfile $fileName] {
	clearsql
	clearoutput .o
	set current(file) [file tail $fileName]
	set result_lines [exec cat -s $fileName]
	.s.sql insert 1.0 "$result_lines"
	setMsg "$fileName loaded"
    } else {
	setMsg "$fileName not found"
    }
    focus .s.sql
}

#
# saveSqlAs
#
#    save the sql code
#

proc saveSqlAs {} {
    global current cmdIdx cmdLast errorCode

    set cmdIdx $cmdLast

    set fileName [tk_getSaveFile -title "TkQuery: Save SQL File" \
	    -filetypes {{"SQL command file" *.sql} {"All Files" *}}]
    if ![string length $fileName] return

    if [catch {open $fileName w} f] {
	setMsg "Error while open '$fileName': [lindex $errorCode 2], not saved"
	return
    }

    set current(file) $fileName

    puts $f [.s.sql get 1.0 end]
    close $f
    setMsg "SQL saved to $current(file)"
}

proc getRowCnt {c} {
    global oldCols

    if !$oldCols($c) {
	return 0
    } else {
	return [expr {int([$c.f.l0 index end])-2}]
    }
}

proc preview {} {
    global options format cmdIdx cmdLast errorCode oldCols tkCompat

    set cmdIdx $cmdLast
    set w      .o
    set c      $w.c
    set sql    .s.sql
    set rows   [getRowCnt $c]

    if !$rows {
	setMsg "No output to preview"
	return
    }
    
    switch $options(val:outputStyle) {
	plain {
	    append results $format(filSta,plain)
	    append results [getSql     $sql plain]
	    append results [getResults $w   plain $rows]
	    append results $format(filEnd,plain)

	    toplevel .preview
	    wm title .preview "TkQuery: Preview"
	    grid rowconfigure    .preview 0 -weight $tkCompat(gridWeightMax)
	    grid columnconfigure .preview 0 -weight $tkCompat(gridWeightMax)

	    button .preview.quit -text Dismiss -com "destroy .preview"
	    grid   .preview.quit -sticky ew -row 2 -columnspan 2

	    scrollbar .preview.xscr -command {.preview.text xview} -orient horiz
	    grid      .preview.xscr -row 1 -column 0 -sticky ew

	    scrollbar .preview.yscr -command {.preview.text yview}
	    grid      .preview.yscr -row 0 -column 1 -sticky ns

	    text .preview.text -wrap none \
		    -xscrollcommand {.preview.xscr set} \
		    -yscrollcommand {.preview.yscr set}
	    if $tkCompat(fontCmd) {
		.preview.text configure -font fixedFont
	    }
	    grid .preview.text -row 0 -column 0 -sticky nsew -ipady 10 -ipadx 10

	    .preview.text insert 1.0 $results
	    .preview.text configure -state disabled
	}

	latex {
	    setMsg "Preview of latex not yet implemented"
	}

	html {
	    set tmpFile "/tmp/tkq[pid].html"
	    if [catch {open $tmpFile w} f] {
		setMsg "Error while open '$tmpFile': [lindex $errorCode 2]"
		return
	    }
	    setMsg "Writing file..."
	    puts $f $format(filSta,html)
	    puts $f [getSql     $sql html]
	    puts $f [getResults $w   html $rows]
	    puts $f $format(filEnd,html)
	    close $f

	    setMsg "Calling netscape..."
	    if [catch "exec netscape -remote openFile($tmpFile)"] {
		tk_messageBox -title "No Netscape running" \
			-message "Since no netscape is running,\
 the html document cannot be previewed" -icon error -type ok
	    }
	    setMsg
	    after 10000 "exec rm $tmpFile > /dev/null 2> /dev/null"
	}
    }
}

#
# saveOut
#
#    save the sql results
#

proc saveOut {{fileName ""}} {
    global cmdIdx cmdLast oldCols errorCode options format

    set w      .o
    set c      $w.c
    set sql    .s.sql
    set cmdIdx $cmdLast
    set style  $options(val:outputStyle)
    set rows   [getRowCnt $c]

    if !$rows {
	setMsg "No output to save"
	return
    }

    if [string length $fileName] {
	set mode   a+
	set append 1
    } else {
	switch $style {
	    html    {set mask *.html}
	    latex   {set mask *.tex}
	    default {set mask *.prt}
	}
	set fileName [tk_getSaveFile -title "TkQuery: Save result" \
		-filetypes {{"Protocol Files" *.prt} {"All Files" *}}]
	if ![string length $fileName] return
	
	set mode   w
	set append 0
    }
    if [catch {open $fileName $mode} f] {
	setMsg "Error while open '$fileName': [lindex $errorCode 2], not saved"
	return
    }

    setMsg "Writing file..."
    if !$append {puts $f $format(filSta,$style)}
    puts $f [getSql     $sql $style]
    puts $f [getResults $w   $style $rows]
    if !$append {puts $f $format(filEnd,$style)}
    close $f

    if !$append {
	setMsg "Results saved to $fileName"
    } else {
	setMsg
    }
}

proc getSql {sql style} {
    global format

    append res "$format(vrbSta,$style)\n"
    append res [$sql get 1.0 end]
    append res "$format(vrbEnd,$style)\n"
    return $res
}

proc getResults {w style rows} {
    global fmt oldCols format options colNames boxConf

    set c $w.c
    append res "[$format(header,$style) fmt $oldCols($c)]\n"
    append res "$format(hdrSta,$style)"
    set currSep ""
    for {set j 0} {$j < $oldCols($c)} {incr j} {
	if $boxConf(visible,$w,$j) {
	    append res $currSep
	    set currSep $format(hdrSep,$style)
	    append res [$format(filter,$style) $fmt($j) [lindex $colNames $j]]
	}
    }
    append res "$format(hdrEnd,$style)\n"

    for {set i 1} {$i <= $rows} {incr i} {
	if {$i > 1} {append res "$format(rowSep,$style)\n"}

	set currSep ""
	for {set j 0} {$j < $oldCols($c)} {incr j} {
	    if $boxConf(visible,$w,$j) {
		append res $currSep
		set currSep $format(colSep,$style)
		append res [$format(filter,$style) $fmt($j) \
			[$c.f.l$j get $i.0 "$i.0 lineend"]]
	    }
	}
	if {$i&&!($i%20)} {setMsg "Writing file ($i rows out of $rows so far)..."}
    }
    append res "$format(tailer,$style)"

    setMsg
    return $res
}

#
# saveSql
#
#    save the sql code to currentFile or use saveSqlAs
#

proc saveSql {} {
    global current cmdIdx cmdLast

    set cmdIdx $cmdLast

    if ![string length $current(file)] {
	saveSqlAs
	return
    }

    set f [open $current(file) w]
    puts $f [.s.sql get 1.0 end]
    close $f
    setMsg "saved to $current(file)"
}

# doNew
#   clear windows
proc doNew {} {
    global current cmdIdx cmdLast

    set cmdIdx $cmdLast

    clearoutput .o
    clearsql
    setMsg
    focus .s.sql
}

# insSql
#   insert the current Sql into the cmdRing
proc insSql {} {
    global cmdRing cmdIdx cmdLast options

    set currentSql [.s.sql get 1.0 end]

    # don't save null buffers
    if ![string length [string trim $currentSql]] return

    set cmdRing($cmdLast) $currentSql
    set cmdIdx  $cmdLast

    if {[incr cmdLast] > $options(val:history)} {
	set cmdLast 0
    }
}

# prevSql
#   save current sql window, replace with previous (dir=-1) or next (dir=1) 
proc prevSql {dir} {
    global cmdRing cmdIdx cmdLast options

    set i 0
    set result_lines ""

    while {$i < $options(val:history) && ![string length $result_lines]} {
	incr cmdIdx $dir

	if {$cmdIdx < 0} {
	    set cmdIdx [expr {$options(val:history)-1}]
	} elseif {$cmdIdx >= $options(val:history)} {
	    set cmdIdx 0
	}
	set result_lines $cmdRing($cmdIdx)
	incr i
    }

    if [string length $result_lines] {
	.s.sql delete 1.0 end
	.s.sql insert 1.0 "$result_lines"
    }
}

proc setExecute {onOff} {
    global execCmd menu stopFlag options

    if $onOff {
	set execCmd "set contFlag 0; set stopFlag 0"
	confMenueEntry $menu(exec) "Execute"  -state disabled
	confMenueEntry $menu(exec) "Cancel"   -state normal
	confMenueEntry $menu(exec) "Continue" -state disabled
	if $options(val:withToolBar) {
	    confToolButton .b exec stop "Interrupt query" \
		    "Interrupt the current sql statement"
	}
    } else {
	set execCmd doSql
	confMenueEntry $menu(exec) "Execute" -state normal
	confMenueEntry $menu(exec) "Cancel"  -state disabled
	if !$stopFlag {
	    confMenueEntry $menu(exec) "Continue" -state normal
	}
	if $options(val:withToolBar) {
	    confToolButton .b exec exec "Execute SQL" \
		    "Execute the SQL statement"
	}
    }
}

# doSql: this is a wrapper proc around trySql,
#        that resets in every case the execution button.

proc doSql {} {
    setMsg
    set currPos 1.0
    set ret     0
    while {[string length $currPos] && !$ret} {
	set ret [catch "trySql $currPos" msg]
	if !$ret {set currPos $msg}
    }

    if $ret {
	global errorInfo
	return -code error -errorinfo $errorInfo $msg
    } else {
	return $msg
    }
}

proc trySql {startPos} {
    global cur adamsg contFlag stopFlag fmt oldCols menu indexes options \
	    protFilename colNames parameter boxConf MASSCNT ROW_NOT_FOUND

    set contFlag 1
    set stopFlag 1

    set sql .s.sql
    getIndexesInSqlWindow $sql $startPos
    set sqlFilt [$sql get $indexes(sqlStart) $indexes(sqlEnd)]
    if ![string length [string trim $sqlFilt]] {
	if ![string compare $indexes(startPos) 1.0] {
	    setMsg "No SQL to execute"
	}
	return ""
    }

    if {[string length $indexes(layoutStart)] \
	    && [string length $indexes(layoutEnd)]} {
	if ![analyzeLayout [$sql get $indexes(layoutStart)+1line \
		                     $indexes(layoutEnd)]] {
	    setMsg "Execution cancelled"
	    return ""
	}
	if [info exists parameter] {
	    foreach parmNo [array names parameter] {
		regsub -all "\[&%]$parmNo" $sqlFilt $parameter($parmNo) sqlFilt
	    }
	}
    }

    insSql
    setMsg "Running SQL"

    set error [catch {callSql $cur $sqlFilt} rc]
    if $error {
	if {[info exists adamsg(rc)] && [info exists adamsg(errortxt)]} {
	    setMsg "Error $adamsg(rc): $adamsg(errortxt)"
	} else {
	    setMsg "Error: $rc"
	}
	if [info exists adamsg(errorpos)] {
	    $sql mark set insert $indexes(sqlStart)+$adamsg(errorpos)chars-1chars
	    $sql see insert
	}
	return ""
    }
    if [info exists adamsg(intoVars)] {
	setMsg "Sorry, but SELECT INTOS not yet supported by TkQuery"
	set thatsAll 1
    } elseif $adamsg(rows) {
	setMsg "SQL finished, $adamsg(rows) rows affected"
	set thatsAll 1
    } elseif ![llength [set colNames [adacols $cur]]] {
	setMsg
	set thatsAll 1
    } else {
	setMsg "SQL finished, getting results"
	setExecute 1
	set thatsAll 0
    }
    if $thatsAll {
	return $indexes(multiSql)
    }

    clearoutput .o 1

    catch {unset fmt}
    if {$adamsg(rc) != $ROW_NOT_FOUND} {
	confMenueEntry $menu(out) "Configure..." -state normal
	confListbox .o [set colNo 0] <start>
	foreach col $colNames {
	    setBoxConf  .o $colNo $col
	    confListbox .o $colNo -textvariable boxConf(title,.o,$colNo)
	    incr colNo
	}
	confListbox .o $colNo <stop>
	if [winfo exists .panel] {mkPanel .panel .o}

	set rows [adafetch $cur -count $MASSCNT]
    }
    set lastnext $adamsg(rc)
    if ![string length $rows] {
	set contFlag 0
    }

    if [string length $indexes(reportStart)] {
	set reportLines [$sql get "$indexes(reportStart)+1line linestart" \
				  "$indexes(reportEnd)"]
	foreach curLine [split $reportLines "\n"] {
	    if {[regexp -nocase "^\[ \t]*NAME\[ \t]*(\[0-9]*)(.*)$" \
		    $curLine match colNo colName]} {
		if {$colNo > 0 && $colNo <= $oldCols(.o.c)} {
		    set colName [string trim $colName]
		    if ![regexp "^'(.*)'$" $colName match colName] {
			set colName [string toupper $colName]
		    }
		    set boxConf(title,.o,[incr colNo -1]) $colName
		}
	    } elseif {[regexp -nocase \
		    "^\[ \t]*(EX|IN)CLUDE\[ \t]*(\[0-9]*)\[ \t]*$"\
		    $curLine match exIn colNo]} {
		if {$colNo > 0 && $colNo <= $oldCols(.o.c)} {
		    switchPanel .o [incr colNo -1] visible \
			    [regexp -nocase IN $exIn]
		}
	    }
	}
    }

    fetchRows $rows

    return $indexes(multiSql)
}

proc getIndexesInSqlWindow {sql startPos} {
    global indexes

    catch {unset indexes}

    set selectionRanges [$sql tag ranges sel]
    if [llength $selectionRanges] {
	set beg [lindex $selectionRanges 0]
	set end [lindex $selectionRanges 1]
    } else {
	set beg $startPos
	set end end
    }
    set indexes(startPos) $beg

    set indexes(multiSql) [$sql search -regexp {^ *[/*]} $beg $end]
    if [string length $indexes(multiSql)] {
	set end $indexes(multiSql)
	append   indexes(multiSql) +1line
    }

    set indexes(layoutStart) \
	    [$sql search -nocase -regexp {^ *LAYOUT}    $beg $end]
    set indexes(layoutEnd) \
	    [$sql search -nocase -regexp {^ *ENDLAYOUT} $beg $end]
    set indexes(reportStart) \
	    [$sql search -nocase -regexp {^ *REPORT}    $beg $end]

    if [string length $indexes(layoutEnd)] {
	set indexes(sqlStart) $indexes(layoutEnd)+1line
    } else {
	set indexes(sqlStart) $beg
    }
    if [string length $indexes(reportStart)] {
	set indexes(sqlEnd)    $indexes(reportStart)
	set indexes(reportEnd) $end
    } else {
	set indexes(sqlEnd) $end
    }
}

proc fetchRows {{rows {}}} {
    global adamsg contFlag stopFlag fmt options cur indexes MASSCNT cnt

    set sql .s.sql

    if [llength $rows] {
	set cnt 0
    } else {
	setMsg "continuing after $cnt rows..."
	set rows [adafetch $cur -count $MASSCNT]
	set contFlag 1
	set stopFlag 1
	setExecute 1
    }

    while {!$adamsg(rc) && $contFlag}  {
	confListbox .o 0 <normal>
	set ix 0
	while {[string length [set row [lindex $rows $ix]]]} {
	    set colNo      0
	    set maxColRows 1
	    foreach col $row {
		switch -glob $fmt($colNo) {
		    LBYTE {
			if [string compare $col $adamsg(nullvalue)] {
			    set col <BYTE>
			}
		    }
		    L* {
			if [string compare $col $adamsg(nullvalue)] {
			    set col [subst -nocommands -novariables \
				          [adareadlong $cur -descriptor $col]]
			}
		    }
		}
		set colList [split $col \n]
		set colRows($colNo) [llength $colList]
		if !$colRows($colNo) {
		    set colRows($colNo) 1
		    insertListbox .o $colNo
		} else {
		    if {$colRows($colNo) > $maxColRows} {
			set maxColRows $colRows($colNo)
		    }
		    foreach colValue $colList {
			insertListbox .o $colNo $colValue
		    }
		}
		incr colNo
	    }
	    if {$maxColRows > 1} {
		for {set iy 0} {$iy < $colNo} {incr iy} {
		    for {set iz $colRows($iy)} {$iz < $maxColRows} {incr iz} {
			insertListbox .o $iy
		    }
		}
	    }
	    incr ix
	    incr cnt
	}

	confListbox .o 0 <disabled>
	updateListboxes .o.c $colNo
	setMsg "$cnt rows so far...."

	# Now allow cancel requests (via menu or stop button) to break through...
	update

	set rows [adafetch $cur -count $MASSCNT]
    }
    updateListboxes .o.c $colNo
    if $options(val:withProt) {
	saveOut $protFilename
    }

    set plural     [expr {$cnt != 1 ? "s" : ""}]
    set terminated [expr {$stopFlag ? "finished" : "interrupted"}]
    if [string length $indexes(multiSql)] {
	setMsg "$cnt row$plural returned, hit any key to continue..."
	set old [bind $sql <Key>]
	bind $sql <Key> "set continueMultiSql 1;break"
	vwait continueMultiSql
	bind $sql <Key> $old
	setMsg
    } else {
	setMsg "SQL $terminated, $cnt row$plural returned"
    }
    setExecute 0
}

# ---------------------------------------------------------------------------
# callSql: This procedure is called anywhere in tkquery as wrapper around 
#          adasql for two reasons:
#          - To protect the environment against malicious select...into, since
#            here adasql is called in a separate namespace (without globals);
#          - if it fails due to a timeout, a reconnect is tried once.
# ---------------------------------------------------------------------------
proc callSql {cur sql args} {
    lappend cmd adasql $cur $sql
    if [llength $args] {eval lappend cmd $args}
    set error [catch $cmd rc]

    if {$error && ![string compare $rc "request failed: connection broken"]} {
	if ![adaReconnect] return
	set error [catch {adasql $cur $sql} rc]
    }
    if $error {return -code error $rc}
    return $rc
}

#
# formatCols
#
#   return a format to use in column printing
#   names, types, and lengths are lists of equal size
#
proc formatCols {names types lengths} {

    while {[llength $names]} {
	set t   [lvarpop types]
	set len [lvarpop lengths]
	set n   [lvarpop names]

	switch $t {
	    fixed - float {set just ""}
	    default       {set just - }
	}

	# make sure length is as long as colunm name 
	set nameLen [string length $n]
	if {$len < $nameLen} {set len $nameLen}

	append fmt "%${just}${len}.${len}s "
    }
    return $fmt
}

proc analyzeLayout {layout} {
    global _layout parameter tkCompat

    catch {unset parameter}
    catch {unset _layout}

    toplevel .layout
    text .layout.t -setgrid 1
    button .layout.dis -text Cancel -command "set _layout 0"
    button .layout.ok  -text Okay   -command "set _layout 1"
    pack .layout.t -fill both -expand 1
    pack .layout.dis .layout.ok -fill x -expand 1 -side left

    if $tkCompat(fontCmd) {
	font create layoutFont -family courier -size 14 -weight bold
	.layout.t configure -font layoutFont
    } else {
	if ![catch {.layout.t configure -font -*-courier-*-r-*-*-14-*-*-*-*-*}] {
	    option add *layout.t*font -*-courier-*-r-*-*-14-*-*-*-*-* widgetDefault
	}
    }

    set maxX     0
    set y        0
    set focusSet 0
    foreach row [split $layout \n] {
	incr y
	.layout.t insert end $row\n
        set wOffset 0
        set cOffset 0
	while {[regexp -indices {([&%][0-9]+)( *)} $row match parmInd blInd]} {
	    set parmStart  [lindex $parmInd 0]
	    set parmEnd    [lindex $parmInd 1]
	    set blankStart [lindex $blInd   0]
	    set blankEnd   [lindex $blInd   1]
	    set fieldEnd   [lindex $blInd   1]
	    if {$blankEnd >= $blankStart} {
		incr fieldEnd -1
	    }

	    # Is there a default value? (e.g. &1 := *)
	    set defFound [regexp -indices "^ *:= *('\[^']*'|\[^ \t]*)" \
		    [string range $row [expr $fieldEnd+1] end] match defInd]
	    if $defFound {
		set defStart  [expr [lindex $defInd 0]+$fieldEnd+1]
		incr fieldEnd [expr [lindex $defInd 1]+1]
		set defString [string range $row $defStart $fieldEnd]
		set defString [string trim $defString ']
	    }

	    set parmNo [string range $row [expr {$parmStart+1}] $parmEnd]
	    if [string length [string index $row [expr {$fieldEnd+1}]]] {
		set parmWidth [expr {$fieldEnd - $parmStart + 1}]
	    } else {
		set parmWidth [expr {80-$cOffset-$parmStart}]
	    }

	    entry .layout.t.p$parmNo -width $parmWidth \
		    -textvariable parameter($parmNo)   \
		    -foreground black -background white -relief sunken
	    if $tkCompat(fontCmd) {
		.layout.t.p$parmNo configure -font layoutFont
	    }
	    bind .layout.t.p$parmNo <F3>     ".layout.dis invoke"
	    bind .layout.t.p$parmNo <Return> ".layout.ok  invoke"
	    .layout.t delete $y.$parmStart+${wOffset}chars \
		             $y.$fieldEnd+${wOffset}chars+1char
	    .layout.t window create $y.$parmStart+${wOffset}chars \
		    -window .layout.t.p$parmNo

	    # Insert the default value, if one was found...
	    if $defFound {
		.layout.t.p$parmNo insert end $defString
	    }

	    if !$focusSet {set focusSet 1; focus .layout.t.p$parmNo}

	    incr wOffset $parmStart
	    incr cOffset $blankEnd
	    set row [string range $row $fieldEnd end]
	}
    }
    .layout.t configure -width 82 -height [incr y] -state disabled

    vwait   _layout
    destroy .layout
    if $tkCompat(fontCmd) {
	font delete layoutFont
    }

    return $_layout
}

#
# showFields
#
#   create a toplevel window with a table's fields
#
proc showFields {argList} {
    global cur adamsg

    set owner [lindex $argList 0]
    set tab   [lindex $argList 1]
    if {![string length $owner] || ![string length $tab]} return

    callSql $cur "SELECT columnname, datatype, len, dec, mode
	            FROM columns
	            WHERE tablename = '$tab'
		      AND owner     = '$owner'
	            ORDER BY pos" -sqlmode adabas

    adafetch $cur -sqlmode adabas -command {
	set name @1
	set t    @2
	set l    @3
	set d    @4
	set m    @5

	switch $t {
	    DATE - TIME - TIMESTAMP - BOOLEAN - INTEGER - SMALLINT {
		set type $t
	    }
	    FLOAT {
		set type ${t}($l,$d)
	    }
	    default {
		set type ${t}($l)
	    }
	}
	switch $m {
	    KEY {set mod KEY}
	    OPT {set mod ""}
	    MAN {set mod "NOT NULL"}
	}
	lappend plist [list $name $type $mod]
    }

    if ![llength $plist] {
	setMsg "No fields in table $owner.$tab"
	return
    }
    set win .${owner}:${tab}
    regsub -all {\$} .${owner}:${tab} _ win
    pickList $win "Columnname Type Mod" $plist "" filtCols
}


#
# showTables: create a toplevel window with user accesable tables
#
proc showTables {} {
    global cur ownerCond

    set query "SELECT owner, tablename
                 FROM tables
                 WHERE $ownerCond
                 ORDER BY owner, tablename"

    if [catch {callSql $cur $query -sqlmode adabas}] {
	setMsg "No user tables available"
	return
    }
    adafetch $cur -command "lappend plist @0" -sqlmode adabas
    pickList .Tables [adacols $cur] $plist showFields filtTab
}


#
# showViewtxt: create a toplevel window with a view's text
#
proc showViewtxt {argList} {
    global cur

    set owner [lindex $argList 0]
    set tab   [lindex $argList 1]

    callSql $cur "SELECT definition
                    FROM viewdefs
                    WHERE viewname = '$tab'
                      AND owner    = '$owner'" -sqlmode adabas

    set longDesc [lindex [adafetch $cur -sqlmode adabas] 0]
    set def      [adareadlong $cur -descriptor $longDesc]

    set win .${owner}:${tab}
    regsub -all {\$} .${owner}:${tab} _ win
    showList $win Definition 550x300 $def
}

#
# showViews: create a toplevel window with user accesable views
#
proc showViews {} {
    global cur ownerCond

    set query "SELECT owner, viewname
                 FROM views
                 WHERE $ownerCond
                 ORDER BY owner, viewname"

    if [catch {callSql $cur $query -sqlmode adabas}] {
	setMsg "No user views available"
	return
    }
    adafetch $cur -command "lappend plist @0" -sqlmode adabas
    adacols  $cur
    pickList .Views "Owner Views" $plist showViewtxt filtTab
}

#
# showIndtxt: create a toplevel window with a index fields
#
proc showIndtxt {argList} {
    global cur connect

    set owner [lindex $argList 0]
    set obj   [lindex $argList 1]
    set tab   [lindex $argList 2]

    set query "SELECT refcolumnname, type
                 FROM $connect(indcoluse)
                 WHERE defindexname = '$obj'
                   AND defowner     = '$owner'
                   AND deftablename = '$tab'
                 ORDER BY pos"

    if [catch {callSql $cur $query -sqlmode adabas}] {
	setMsg "No index columns in index $owner.$tab"
	return
    }
    adafetch $cur -command "lappend plist @1; set uniq @2" -sqlmode adabas

    set win .${owner}:${tab}
    regsub -all {\$} .${owner}:${tab} _ win
    pickList $win [list "Columns of [string tolower $uniq] index $obj"] $plist "" 
}

#
# showIndexes: create a toplevel window with user accesable indexes
#
proc showIndexes {} {
    global cur ownerCond

    set query "SELECT owner, indexname, tablename
                 FROM indexes
                 WHERE $ownerCond
                 ORDER BY owner, indexname, tablename"

    if [catch {callSql $cur $query -sqlmode adabas}] {
	setMsg "No user indexes available"
	return
    }
    adacols  $cur
    adafetch $cur -command "lappend plist @0" -sqlmode adabas
    pickList .Indexes [list Owner Index "on table"] $plist showIndtxt
}

proc aboutHelp {} {
    tk_messageBox -title "About TkQuery" -message "Adabas D: TkQuery
Version 0.6\n\nJanuary 1997\n\nChristian Krone
krischan@cs.tu-berlin.de" -type ok
}

########################
#
# pickList
#
#   return a selection from a listbox by calling a proc
#

proc pickList {win heading plist callproc {filtproc NoFilterProc}} {

  set win_title $win
  regsub -all {:} $win_title .   win_title

  set win [string tolower $win]
  
  catch {destroy $win}
  toplevel $win
  wm title $win [string range $win_title 1 end]

  # find if callproc is a valid command
  set doproc [string length [info commands $callproc]]
 
  # try to place window away from the main toplevel
  regexp {\+([0-9]*)\+([0-9]*)} [winfo geom .] match x y
  wm geom $win +[incr x 50]+[incr y 150]

  pannedListboxes $win.o
  canvasConf $win.o.c
  confListbox $win.o [set colNo 0] <start>
  foreach col $heading {
      setBoxConf  $win.o $colNo $col
      confListbox $win.o $colNo -text $col
      if $doproc {
	  $win.o.c.f.l$colNo configure -cursor hand2
	  bind $win.o.c.f.l$colNo <1> "$callproc \[listboxGetLine $win.o $colNo]"
      }
      incr colNo
  }
  confListbox $win.o $colNo <stop>

  confListbox $win.o 0 <normal>
  foreach row $plist {
      set colNo 0
      foreach cell $row {
	  insertListbox $win.o $colNo $cell
	  incr colNo
      }
  }
  confListbox $win.o 0 <disabled>

  button $win.can -text "Cancel" -relief raised -borderwidth 2 \
		-command "destroy $win"
  pack $win.can  -side bottom -fill x -expand 1
}

proc showList {win heading geom plist} {
    global font tkCompat

    set win_title $win
    regsub -all {:} $win_title . win_title

    set win [string tolower $win]
  
    catch {destroy $win}
    toplevel $win
    wm title $win [string range $win_title 1 end]

    # try to place window away from the main toplevel
    set topgeom [split [split [winfo geom .] x] +]
    set newx [expr {[lindex $topgeom 1] +  50 } ]
    set newy [expr {[lindex $topgeom 2] + 150 } ]
    wm geom $win ${geom}+${newx}+$newy
    set w [lindex [split $geom x] 0]
    set h [lindex [split $geom x] 1]
    wm minsize $win 120 120

    frame $win.l 
    frame $win.f 

    label $win.l.l -text $heading -anchor w
    scrollbar $win.f.vert -orient vertical -command "$win.f.box yview" \
	    -relief sunken
    text $win.f.box -yscroll "$win.f.vert set" -relief sunken -wrap word
    if $tkCompat(fontCmd) {
	$win.f.box configure -font outFont
    } else {
	$win.f.box configure -font $font
    }

    button $win.can -text "Cancel" -relief raised -borderwidth 2 \
	    -command "destroy $win"

    $win.f.box insert 1.0 $plist

    pack $win.l.l    -side top -fill x -anchor nw
    pack $win.f.vert -side right -fill both
    pack $win.f.box  -side left -fill both -expand 1

    pack $win.can -side bottom -fill x -expand 1
    pack $win.l   -fill x
    pack $win.f   -fill both -expand 1

    bind $win <Enter> "focus $win.f.box"
    bind $win.f.box <F3> "$win.b.can invoke"
}


proc setNull {} {
    global adamsg

    set oldVal $adamsg(nullvalue)
    if ![topEntry .snull "Set Null Value" Value adamsg(nullvalue)] {
	set adamsg(nullvalue) $oldVal
	return
    }
    set adamsg(nullvalue) [string trim $adamsg(nullvalue)]
}

proc setSpecialNull {} {
    global adamsg

    set oldVal $adamsg(specialnull)
    if ![topEntry .snull "Set Special Null" Value adamsg(specialnull)] {
	set adamsg(specialnull) $oldVal
	return
    }
    set adamsg(specialnull) [string trim $adamsg(specialnull)]
}

proc doSwitch {} {
    global lda cur switchReady switch

    toplevel .sw
    wm title .sw "TkQuery: Switch"

    frame .sw.ak  -relief raised -borderwidth 2
    frame .sw.bd  -relief raised -borderwidth 2
    frame .sw.kb  -relief raised -borderwidth 2
    frame .sw.ad  -relief raised -borderwidth 2
    frame .sw.but -relief raised -borderwidth 2

    label .sw.ak.akhead -text "AK debugging"
    label .sw.bd.bdhead -text "BD debugging"
    label .sw.kb.kbhead -text "KB debugging"
    label .sw.ad.adhead -text "Stored procedures"

    checkbutton .sw.ak.ak -text "AK  ak; not AI, AJ"  -var switch(ak) -an w
    checkbutton .sw.ak.ac -text "AC  ak_cache"        -var switch(ac) -an w
    checkbutton .sw.ak.ai -text "AI  ak_isolev"       -var switch(ai) -an w
    checkbutton .sw.ak.aj -text "AJ  ak_join"         -var switch(aj) -an w
    checkbutton .sw.ak.as -text "AS  ak_sem"          -var switch(as) -an w
    checkbutton .sw.ak.at -text "AT  ak_strat"        -var switch(at) -an w
    checkbutton .sw.ak.ay -text "AY  ak_syn"          -var switch(ay) -an w
    label       .sw.ak.e1 -text ""
    checkbutton .sw.ak.fh -text "FH  fs_ak"           -var switch(fh) -an w
    checkbutton .sw.ak.tp -text "TP  test_ak"         -var switch(tp) -an w
    label       .sw.ak.e2 -text ""
    label       .sw.ak.e3 -text ""
    checkbutton .sw.ak.tb -text "TB  test_bd"         -var switch(tb) -an w

    checkbutton .sw.bd.bd -text "BD  bd; not BI"      -var switch(bd) -an w
    checkbutton .sw.bd.bb -text "BB  bd_buf"          -var switch(bb) -an w
    checkbutton .sw.bd.by -text "BY  bd_byte"         -var switch(by) -an w
    checkbutton .sw.bd.bx -text "BX  bd_index"        -var switch(bx) -an w
    checkbutton .sw.bd.bv -text "BV  bd_inv"          -var switch(bv) -an w
    checkbutton .sw.bd.bo -text "BO  bd_io"           -var switch(bo) -an w
    checkbutton .sw.bd.bq -text "BQ  bd_ioqueue"      -var switch(bq) -an w
    checkbutton .sw.bd.bm -text "BM  bd_keym"         -var switch(bm) -an w
    checkbutton .sw.bd.bl -text "BL  bd_lock"         -var switch(bl) -an w
    checkbutton .sw.bd.bf -text "BF  bd_oflw"         -var switch(bf) -an w
    checkbutton .sw.bd.bp -text "BP  bd_psm"          -var switch(bp) -an w
    checkbutton .sw.bd.bt -text "BT  bd_tbuf"         -var switch(bt) -an w
    checkbutton .sw.bd.bi -text "BI  <bd_interface>"  -var switch(bi) -an w

    checkbutton .sw.kb.kb -text "KB  kb; not KE"      -var switch(kb) -an w
    checkbutton .sw.kb.kd -text "KD  kb_dist"         -var switch(kd) -an w
    checkbutton .sw.kb.kf -text "KF  kb_func"         -var switch(kf) -an w
    checkbutton .sw.kb.kl -text "KL  kb_lock"         -var switch(kl) -an w
    checkbutton .sw.kb.ke -text "KE  kb_lockentry"    -var switch(ke) -an w
    checkbutton .sw.kb.kg -text "KG  kb_log"          -var switch(kg) -an w
    checkbutton .sw.kb.kn -text "KN  kb_net, kb_dist" -var switch(kn) -an w
    checkbutton .sw.kb.kq -text "KQ  kb_qual"         -var switch(kq) -an w
    checkbutton .sw.kb.ks -text "KS  kb_sel"          -var switch(ks) -an w
    checkbutton .sw.kb.kv -text "KV  kb_save"         -var switch(kv) -an w
    label       .sw.kb.e1 -text ""
    checkbutton .sw.kb.fl -text "FL  fs_kb"           -var switch(fl) -an w
    checkbutton .sw.kb.tk -text "TK  test_kb"         -var switch(tk) -an w

    checkbutton .sw.ad.pr -text "PR"                  -var switch(tk) -an w
    checkbutton .sw.ad.pi -text "PI  sproc"           -var switch(pi) -an w
    checkbutton .sw.ad.pc -text "PC  sproc_call"      -var switch(pc) -an w
    checkbutton .sw.ad.ps -text "PS  sproc_sql"       -var switch(ps) -an w
    label       .sw.ad.e1 -text ""
    label       .sw.ad.h2 -text "Other functions"
    checkbutton .sw.ad.gg -text "GG"                  -var switch(gg) -an w
    checkbutton .sw.ad.st -text "ST  st_stack"        -var switch(st) -an w
    checkbutton .sw.ad.uc -text "UC  unicode"         -var switch(uc) -an w

    frame  .sw.but.default -relief sunken -bd 1
    button .sw.but.sw -text Switch -command {set switchReady 1; destroy .sw}
    button .sw.but.re -text Reset  -command {
	foreach currSw [array names switch] {
	    set switch($currSw) 0
	}
    }
    button .sw.but.vt -text Vtrace -command {set switchReady 2; destroy .sw}
    button .sw.but.ca -text Cancel -command {set switchReady 0; destroy .sw}
    bind .sw.but.sw <Return> ".sw.but.sw invoke"
    pack   .sw.but.default .sw.but.re .sw.but.vt .sw.but.ca \
	    -padx 10 -pady 10 -ipadx 10 -fill x
    pack   .sw.but.sw -in .sw.but.default -padx 1 -pady 1 -fill both -expand 1

    pack .sw.ak .sw.bd .sw.kb .sw.ad .sw.but -side left -fill both -expand 1

    pack .sw.ak.akhead .sw.ak.ak .sw.ak.ac .sw.ak.ai .sw.ak.aj .sw.ak.as \
	    .sw.ak.at .sw.ak.ay .sw.ak.e1 .sw.ak.fh .sw.ak.tp .sw.ak.e2  \
	    .sw.ak.e3 .sw.ak.tb -side top -anchor w -fill x
    pack .sw.bd.bdhead .sw.bd.bd .sw.bd.bb .sw.bd.by .sw.bd.bx .sw.bd.bv \
	    .sw.bd.bo .sw.bd.bq .sw.bd.bm .sw.bd.bl .sw.bd.bf .sw.bd.bp  \
	    .sw.bd.bt .sw.bd.bi -side top -anchor w -fill x
    pack .sw.kb.kbhead .sw.kb.kb .sw.kb.kd .sw.kb.kf .sw.kb.kl .sw.kb.ke \
	    .sw.kb.kg .sw.kb.kn .sw.kb.kq .sw.kb.ks .sw.kb.kv .sw.kb.e1  \
	    .sw.kb.fl .sw.kb.tk -side top -anchor w -fill x
    pack .sw.ad.adhead .sw.ad.pr .sw.ad.pi .sw.ad.pc .sw.ad.ps .sw.ad.e1 \
	    .sw.ad.h2 .sw.ad.gg .sw.ad.st .sw.ad.uc                      \
	    -side top -anchor w -fill x

    focus .sw.but.sw
    vwait switchReady
    switch $switchReady {
	1 {
	    set debug ""
	    foreach currSw [array names switch] {
		if $switch($currSw) {
		    lappend debug $currSw
		}
	    }
	    set layer AKBD
	    if [catch {adaspecial $lda switch $layer $debug} msg] {
		setMsg $msg
	    }
	}
	2 {
	    adasql $cur VTRACE -sqlmode adabas
	}
    }
}

proc lvarpop {list} {
    upvar 1 $list x

    set first [lindex $x 0]
    set x [lreplace $x 0 0]
    return $first
}

proc max {a b} {
    if {$a > $b} {return $a} {return $b}
}

proc makeTransient {w} {
    wm transient $w .
    set xpos [expr [winfo rootx .]+[winfo width .]/3]
    set ypos [expr [winfo rooty .]+[winfo height .]/3]
    wm geom $w +${xpos}+$ypos
}

proc confMenueEntry {m index args} {
    set tearOffs [info commands .tearoff*]
    if [llength $tearOffs] {
        set m "$m $tearOffs"
    }
    foreach menu $m {
        catch {eval $menu entryconfigure [list $index] $args}
    }
}

# ---------------------------------------------------------------------
# 'adaConnect' will be called, when the user presses the Connect button
# in the connect dialog. It tries to connect with the given user/password
# combination to the database, and sets - if succesful - the global
# variables 'lda' (login handle) and 'cur' (cursor).
# ---------------------------------------------------------------------
proc adaConnect {} {
    global connect lda cur

    if ![tryConnect] {return 0}

    set lda $connect(db)
    set cur [adaopen $lda]

    return 1
}

proc adaReconnect {} {
    global cur lda options adamsg

    catch {adaclose  $cur}
    catch {adalogoff $lda}

    if ![adaConnect] {
	catch {unset adamsg(errortxt)}
	setMsg "Connection to database server broken"
	return 0
    }

    initOptions .o.c
    if !$options(val:autoCommit) {
	setMsg "Work rolled back (inactivity timeout)"
    }
    return 1
}

proc initOptions {c} {
    global adamsg options

    setAutoCommit 1
    setFormat     1
    setProt
    set adamsg(nullvalue)   $options(val:nullValue)
    set adamsg(specialnull) $options(val:specialNull)
}

# ---------------------------------------------------------------------
# Mainprogram
# ---------------------------------------------------------------------
wm withdraw .
loadAdabastcl
getDbnameUserPassword "Adabas D: TkQuery" adaConnect
createMain
