/*
 * port8.h --
 *
 * This file contains macros, that are neccessary only on some specific
 * architectures. It also defines macros that make it possible, to use
 * the new (from version 8.0 upward) object commands of Tcl, even
 * with a Tcl version, that is older.
 *
 * Copyright (c) 1996-1997 Christian Krone. All rights reserved.
 * This program is free software; you can redistribute it and/or
 * modify it under the same terms as Tcl itself.
 * See also licence.terms
 *
 * $Id: AdabasPort.h,v 1.4 1997/01/01 21:49:36 adabas Exp adabas $
 */

/*
 * If we are on a Windows system, we have to set some defines here,
 * since there exists no configure.
 */

#ifdef __WIN32__
#  if TCL_MAJOR_VERSION >= 8
#  define HAS_TCL_OBJECTS 1
#  endif
#endif

/*
 * Not all systems have unistd.h (e.g. Windows).
 */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#else
   EXTERN long lseek _ANSI_ARGS_((int fd, long offset, int whence));
   EXTERN int  read  _ANSI_ARGS_((int fd, char *buf, size_t size));
   EXTERN int  write _ANSI_ARGS_((int fd, char *buf, size_t size));
   EXTERN int  close _ANSI_ARGS_((int fd));
#endif

/*
 * For calls of the procedure open on windows32 systems the flag O_RAW,
 * which permits end of line translations, must be or'ed to the mode argument.
 * On systems, where this flag isn't defined, we are using 0 (noop) instead.
 */

#ifndef O_RAW
#define O_RAW 0
#endif

/*
 * Backward compatibility to Tcl7.x.
 *
 * It is better to ask an extra identified here instead of asking the
 * Tcl version directly, since so we can (for test purpose) switch to
 * the old behavior even with a new Tcl version. This identifier has
 * to be set by configure (disabled with --without-tclobj).
 */

#ifdef HAS_TCL_OBJECTS

#define SetResult(interp,str) \
     (Tcl_SetStringObj (Tcl_GetObjResult (interp), (str), -1))
#define AppendResult(interp,str) \
     (Tcl_StringObjAppend ((interp), Tcl_GetObjResult (interp), (str), -1))
#define SetIntResult(interp,val) \
     (Tcl_SetIntObj (Tcl_GetObjResult (interp), (val)))

#else

#define Tcl_ObjCmdProc Tcl_CmdProc
#define Tcl_CreateObjCommand(interp,name,length,proc,data,delete) \
     Tcl_CreateCommand(interp,name,proc,data,delete)

#define objc    argc
#define objv    argv
#define Tcl_Obj char

#define Tcl_GetStringFromObj(interp,obj,lenPtr) \
	((*(lenPtr) = strlen(obj)), (obj))
#define Tcl_GetBooleanFromObj Tcl_GetBoolean
#define Tcl_ResetObjResult Tcl_ResetResult
#define SetResult(interp,str) ((interp)->result = (str))
#define AppendResult(interp,str) \
      (Tcl_AppendResult ((interp), (str), (char *) NULL))

#define Tcl_GetIntFromObj GetIntFromObj

EXTERN int GetIntFromObj _ANSI_ARGS_((Tcl_Interp *interp, Tcl_Obj *obj,
				      long *longPtr));
EXTERN int SetIntResult _ANSI_ARGS_((Tcl_Interp *interp, long val));

#endif
